/* erandom.c : a provably secure pseudorandom number generator
   copyright 2003, seth hardy: shardy@tsumego.com
	 tsumego foundation, 2193 commonwealth ave. suite 336, boston MA 02135 USA
	
	 acknowledgements:
	
	 quite a bit of this code, especially the erandom_poll() and erandom_read()
   functions,	was taken from random.c, courtesy of theodore ts'o. thanks!	 
	 
	 the general setup and some skeleton code was taken from "linux device
	 drivers", 2nd edition, by alessandro rubini and jonathan corbet. thanks! i 
	 doubt i'd have been able to write kernel code without your book.
	 
	 --
	 This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
	 
	 This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	 
*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/poll.h>

#define AUTHOR      "Seth Hardy"
#define DESCRIPTION "A provably secure prng for Linux."
#define LICENSE     "GPL"
#define DEVMAJOR		123
#define DEVNAME			"erandom"
#define STORESIZE   1024 * 16

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);
MODULE_SUPPORTED_DEVICE(DEVNAME);

// definition of the bytestore, to hold good/bad/output bits
struct bytestore {
	unsigned int size;
  unsigned int threshold;   // parameter based on the extractor core
	__u8 offset;
  __u8 *bytes;
	volatile int lock;
};

typedef void(*extptr)(__u8 *, __u8 *, __u8 *);
typedef void(*exportptr)(__u8 *, __u8 *, int, int);
typedef struct bytestore bytestore;
	
// file operations prototypes
ssize_t erandom_read(struct file *, char *, size_t, loff_t *);
ssize_t erandom_write(struct file *, const char *, size_t, loff_t *);
int erandom_open(struct inode *, struct file *);
int erandom_release(struct inode *, struct file *);
unsigned int erandom_poll(struct file *, poll_table *);

struct file_operations erandom_fops = {
	read:			erandom_read,
	//write:		erandom_write,	// not supported yet, but soon...
	open:			erandom_open,
	release:	erandom_release,
	poll:			erandom_poll
	//ioctl:	erandom_ioctl,		// these are next on the list...
};

// exported function prototypes
int register_core(extptr,int,int,int);
int unregister_core(int);

// nonexported function prototypes
int store_add(bytestore *, __u8 *, int);
int store_get(bytestore *, __u8 *, int);
void devrandom_hook(__u8 *, __u8 *, int, int);

// provided by the patch to random.c, provides erandom with entropy
extern int register_erandom(exportptr);
extern int unregister_erandom(void);

// variables
static extptr core = NULL;
static bytestore badbytes;
static bytestore goodbytes;
static bytestore extbytes;
static int hook = 0;

static DECLARE_WAIT_QUEUE_HEAD(erandom_read_wait);
static DECLARE_WAIT_QUEUE_HEAD(store_wait);
static DECLARE_WAIT_QUEUE_HEAD(hook_wait);

/* ------------------------------------------------------- */

void devrandom_hook(__u8 *gbuf, __u8 *bbuf, int gs, int bs) {
// function sent to random.c to send entropy back to erandom
  
	int c;
	
	wait_event_interruptible(hook_wait,(hook == 0));
	hook = 1;
	
  store_add(&goodbytes,gbuf,gs);
  store_add(&badbytes,bbuf,bs);
	
  if( (goodbytes.size >= goodbytes.threshold) &&
      (badbytes.size >= badbytes.threshold) &&
      (extbytes.size < STORESIZE) &&
      (core != NULL)	) {
		
		// eventually get allocation/deallocation out of the hook...
    __u8 *gx, *bx, *ex;
				
    gx = (__u8 *)kmalloc(goodbytes.threshold,GFP_KERNEL);
    bx = (__u8 *)kmalloc(badbytes.threshold,GFP_KERNEL);
    ex = (__u8 *)kmalloc(extbytes.threshold,GFP_KERNEL);
		
		if( (gx == NULL) || (bx == NULL) || (ex == NULL) ) {
			printk(KERN_WARNING "erandom: could not allocate memory\n");
			hook = 0;
			return;
		}
				
		c = 0;
		while( c < goodbytes.threshold)
			c += store_get(&goodbytes,gx+c,goodbytes.threshold-c);
		c = 0;
		while( c < badbytes.threshold) 
			c += store_get(&badbytes,bx+c,badbytes.threshold-c);
	  core(ex,gx,bx);
		store_add(&extbytes,ex,extbytes.threshold*8);
		if(extbytes.size > 0) 
			wake_up_interruptible(&erandom_read_wait);
    // this printk is useful for watching erandom go...
		//printk("bits extracted. %d %d ; %d\n",goodbytes.size,badbytes.size,extbytes.size);		
		
		memset((void *)gx,0,goodbytes.threshold);
		memset((void *)bx,0,badbytes.threshold);
		memset((void *)ex,0,extbytes.threshold);
		
		kfree(gx);
    kfree(bx);
    kfree(ex);
  }
	
	hook = 0;
	
	return;

}

int store_add(bytestore *bs, __u8 *buf, int s) {
  
  int i;
	
	//printk("adding %d bits from store of size %d at offset %d\n",s,bs->size,bs->offset);
	
	if( bs->size == STORESIZE ) return 0;
	
	wait_event_interruptible(store_wait,(bs->lock == 0));
	
	bs->lock = 1;
	
	i = 0;
	
	while( (s >= 8) && (bs->size < STORESIZE-1) ) {
		bs->bytes[bs->size+1] = bs->bytes[bs->size];
		bs->bytes[bs->size] = buf[i];
		i++;
		bs->size++;
		s -= 8;
	}
	
	if( s >= 8 ) {
		// this means bs->size == STORESIZE-1
		bs->bytes[STORESIZE-1] = buf[i];
		bs->size = STORESIZE;
		bs->offset = 0;
		i++;
		bs->lock = 0;
		return i;
	}
	
	if( bs->offset + s <= 8 ) {
		bs->bytes[bs->size] |= (buf[i] << bs->offset);
		bs->offset += s;
		if( bs->offset == 8 ) {
			bs->offset = 0;
			bs->size++;
			i++;
		}
	}
	else {
		bs->bytes[bs->size] |= (buf[i] << bs->offset);
		if( bs->size < STORESIZE - 1 ) { 
			bs->bytes[bs->size+1] = buf[i] >> (8-bs->offset);
		  bs->offset = (bs->offset + s) % 8;
		}
		else bs->offset = 0;
		bs->size++;
		i++;
	}
	
	bs->lock = 0;
	
	return i;
		
}

int store_get(bytestore *bs, __u8 *buf, int s) {
  
	//printk("removing %d bits from store of size %d at offset %d\n",s,bs->size,bs->offset);
	
	if( bs->size == 0 ) return 0;

	wait_event_interruptible(store_wait,(bs->lock == 0));
	
	bs->lock = 1;
	
  if( bs->size < s ) {
		memcpy(buf,bs->bytes,bs->size);
		memset(bs->bytes,0,bs->size);
		if(bs->offset != 0) bs->bytes[0] = bs->bytes[bs->size];
		s = bs->size;
		bs->size = 0;
	}
  else {  
    memcpy(buf,&(bs->bytes[bs->size-s]),s);
    memset(&(bs->bytes[bs->size-s]),0,s);
    if( bs->offset != 0 ) bs->bytes[bs->size-s] = bs->bytes[bs->size];
    bs->size -= s;
  }
	
	bs->lock = 0;
	
  return s;
  
}

int unregister_core(int num) {
  
  // eventually reupdate to handle multiple cores
  // too bad that code was lost, but now I remember to sync the fs before
	//   testing something that can hose the kernel :(
  
  core = NULL;
	goodbytes.threshold = 0;
	badbytes.threshold = 0;
	extbytes.threshold = 0;
	
	MOD_DEC_USE_COUNT;
  
  return 0;
  
}

int register_core(extptr c, int gs, int bs, int es) {
  
	// eventually reupdate to handle multiple cores
 
  core = c;
	goodbytes.threshold = gs;
  badbytes.threshold = bs;
	extbytes.threshold = es;
	
	MOD_INC_USE_COUNT;
	
  return 1;
  
}


int erandom_init(void) {
  
	int reg;
	
	SET_MODULE_OWNER(&erandom_fops);
	
	reg = register_chrdev(DEVMAJOR,DEVNAME,&erandom_fops);
	
	if( reg < 0 ) {
		printk(KERN_WARNING "erandom: unable to register device number %d\n", DEVMAJOR);
		return reg;
  }
	
	core = NULL;
	
	goodbytes.size = 0;
	goodbytes.offset = 0;
  goodbytes.threshold = 0;
	goodbytes.lock = 0;
  goodbytes.bytes = (__u8 *)kmalloc(STORESIZE,GFP_KERNEL);
	memset(goodbytes.bytes,0,STORESIZE);
  
  if( goodbytes.bytes == NULL ) {
    printk(KERN_WARNING "erandom: unable to register core; out of memory\n");
    return -ENOMEM;
  }
  
  badbytes.size = 0;
	badbytes.offset = 0;
  badbytes.threshold = 0;
	badbytes.lock = 0;
  badbytes.bytes = (__u8 *)kmalloc(STORESIZE,GFP_KERNEL);
	memset(badbytes.bytes,0,STORESIZE);
  
  if( badbytes.bytes == NULL ) {
    kfree(goodbytes.bytes);
    printk(KERN_WARNING "erandom: unable to register core; out of memory\n");
    return -ENOMEM;
  }
  
  extbytes.size = 0;
	extbytes.offset = 0;
  extbytes.threshold = 0;
	extbytes.lock = 0;
  extbytes.bytes = (__u8 *)kmalloc(STORESIZE,GFP_KERNEL);
	memset(extbytes.bytes,0,STORESIZE);
  
  if( extbytes.bytes == NULL ) {
    kfree(badbytes.bytes);
    kfree(goodbytes.bytes);
    printk(KERN_WARNING "erandom: unable to register core; out of memory\n");
    return -ENOMEM;
  }
  
	register_erandom(&devrandom_hook);
	
  return 0;
  
}

void erandom_cleanup(void) {
  
  int unreg;
  
  unreg = unregister_chrdev(DEVMAJOR,DEVNAME);
  
  if( unreg < 0 ) 
    printk(KERN_WARNING "erandom: unable to unregister device number %d\n",DEVMAJOR);
  
	memset(goodbytes.bytes,0,STORESIZE);
	goodbytes.size = 0;
  goodbytes.threshold = 0;
	goodbytes.offset = 0;
  kfree(goodbytes.bytes);
  
	memset(badbytes.bytes,0,STORESIZE);
  badbytes.size = 0;
  badbytes.threshold = 0;
	badbytes.offset = 0;
  kfree(badbytes.bytes);
  
	memset(extbytes.bytes,0,STORESIZE);
  extbytes.size = 0;
  extbytes.threshold = 0;
	badbytes.offset = 0;
  kfree(extbytes.bytes);
	
	unregister_erandom();
	
}

int erandom_open(struct inode *inode, struct file *filp) {
	
	MOD_INC_USE_COUNT;
	
	return 0;
	
}

int erandom_release(struct inode *inode, struct file *filp) {
	
	MOD_DEC_USE_COUNT;
	
	return 0;
	
}

unsigned int erandom_poll(struct file *file, poll_table *wait) {
	
	unsigned int mask;

	poll_wait(file, &erandom_read_wait, wait);
	mask = 0;
	if (extbytes.size > 0)
		mask |= POLLIN | POLLRDNORM;
	return mask;
	
}

ssize_t erandom_read(struct file *filp, char *buf, size_t nbytes, loff_t *offp) {
// lots of code in here taken from random.c. thanks!
	
	DECLARE_WAITQUEUE(wait, current);
	ssize_t	n, retval = 0, count = 0;
	
	if (nbytes == 0) return 0;

	add_wait_queue(&erandom_read_wait, &wait);
	while (nbytes > 0) {
		set_current_state(TASK_INTERRUPTIBLE);
		n = nbytes;
		if(n > extbytes.size) n = extbytes.size;
		if (n == 0) {
			if (filp->f_flags & O_NONBLOCK) {
				retval = -EAGAIN;
				break;
			}
			if (signal_pending(current)) {
				retval = -ERESTARTSYS;
				break;
			}
			schedule();
			continue;
		}
		n = store_get(&extbytes,(__u8 *)buf,n);
		if (n < 0) {
			retval = n;
			break;
		}
		count += n;
		buf += n;
		nbytes -= n;
		break;	/* like a named pipe, right right */
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&erandom_read_wait, &wait);

	if (count != 0) UPDATE_ATIME(filp->f_dentry->d_inode);
	
	return (count ? count : retval);
	
}

ssize_t erandom_write(struct file *filp, const char *buf, size_t count, 
                      loff_t *offp) {
  // yeah, this isn't supported yet
  return 0;
}

module_init(erandom_init);
module_exit(erandom_cleanup);

EXPORT_SYMBOL_NOVERS(register_core);
EXPORT_SYMBOL_NOVERS(unregister_core);
