/* k1q-uhfe.c : a (k-1)/q - universal hash family extractor core for erandom
   copyright 2003, seth hardy: shardy@tsumego.com
	 tsumego foundation, 2193 commonwealth ave. suite 336, boston MA 02135 USA	 
	 
	 --
	 This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
	 
	 This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	 
*/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/types.h>

#define AUTHOR      "Seth Hardy"
#define DESCRIPTION "Demo extractor core: (k-1)/q universal hash family."
#define LICENSE     "GPL"

// extractor parameter k:
// (q;q^k;q) hash family which is (k-1)/q universal.
#define K_POWER     9

// field size, in bits
#define FIELD_SIZE  256

// field size, in limbs of 32 bits
#define FIELD_LIMBS FIELD_SIZE/32

// field size, in bytes
#define FIELD_BYTES FIELD_SIZE/8

// input/output sizes, in bytes
#define GOODINPUT 32
#define BADINPUT	288
#define EXTOUTPUT	64

// irreducible polynomial (mod 2): x^256 + x^254 + x^251 + x^246 + 1
// x^256 is implied by the way we use this: x^256 = the rest...
const __u32 IRR_256[] = {0x1,0,0,0,0,0,0,0x48400000};

// the slot in erandom this core is loaded as
int corenum;

// GF(2^x) functions
void gf2x_add(int, __u32 *, __u32 *, __u32 *);
void gf2x_mul(__u32 *, __u32 *, __u32 *, const __u32 *);
void gf2x_shl(int, __u32 *, __u32 *);

// the extractor
void extract(__u8 *, __u8 *, __u8 *);

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

typedef void(*extptr)(__u8 *,__u8 *,__u8 *);
extern int register_core(extptr,int,int,int);
extern int unregister_core(int);


int k1q_uhfe_init(void) {
	
	corenum = register_core(&extract,GOODINPUT,BADINPUT,EXTOUTPUT);
	
	if( corenum < 0 ) {
		printk(KERN_WARNING "k1q-uhfe: unable to load core k1q-uhfe\n");
		return -1;
	}
	
	printk("k1q-uhfe: loaded core k1q-uhfe\n");
	
	return 0;
	
}

void k1q_uhfe_cleanup(void) {
	
	unregister_core(corenum);
	
}


void gf2x_add(int size, __u32 *x, __u32 *a, __u32 *b) {
  
  int i;
  
  for(i=size-1;i>=0;i--) x[i] = a[i] ^ b[i];
  
  return;
  
}

void gf2x_mul(__u32 *x, __u32 *a, __u32 *b, const __u32 *irr) {

  int i,j,k;
  __u32 abit;
  __u32 foo;
  __u32 acc[FIELD_LIMBS*2], stmp[FIELD_LIMBS*2], ttmp[FIELD_LIMBS*2];
	
  memset((void *)acc,0,FIELD_BYTES*2);
	
	// multiply; intermediate result will be double length
  for(i=0;i<FIELD_SIZE;i++) {	
		foo = 1 << i%32;
		abit = (a[i/32] & foo);
    if(abit) {
			gf2x_shl(i,b,stmp);
			gf2x_add(2*FIELD_LIMBS,ttmp,stmp,acc);
			for(j=0;j<FIELD_LIMBS*2;j++) acc[j] = ttmp[j];
		}
  }
	
  // reduce
	for(i=2*FIELD_LIMBS-1;i>=FIELD_LIMBS;i--) {
    for(j=31;j>=0;j--) {
      foo = 1 << j;
      if( (foo & acc[i]) != 0 ) {
				foo = ~foo;
				acc[i] &= foo;
				gf2x_shl((i-FIELD_LIMBS)*32+j,IRR_256,stmp);
				gf2x_add(2*FIELD_LIMBS,ttmp,acc,stmp);
	      for(k=0;k<FIELD_LIMBS*2;k++) acc[k] = ttmp[k];
      }
    }
  }
	
  for(i=0;i<FIELD_LIMBS;i++) x[i] = acc[i];
  
  return;

}

void gf2x_shl(int sh, __u32 *x, __u32 *y) {

  int off = sh % 32;
  int s   = sh / 32;
  int i,j;
	
	__u32 blah;
  
  memset(y,0,FIELD_BYTES*2);
  for(i=0;i<FIELD_LIMBS;i++) y[i+s] = x[i];
	
	// loop done with >>1 because of bug...
  for(i=FIELD_LIMBS*2-1;i>0;i--) {
    y[i] <<= off;
		blah = y[i-1];
		for(j=0;j<32-off;j++) blah >>= 1;
		y[i] |= blah;
  }
	
  y[0] <<= off;
  
  return;
  
}

void extract(__u8 *output, __u8 *good, __u8 *bad) {
	
	int i;
	
	__u32 tmp[FIELD_LIMBS], acc[FIELD_LIMBS], xi[FIELD_LIMBS], ai[FIELD_LIMBS];
	__u32 gd[FIELD_LIMBS];
	
	MOD_INC_USE_COUNT;
	
	memset(ai,0,FIELD_BYTES);
	ai[0] = 1;
	memcpy((void *)acc,(void *)bad,FIELD_BYTES);
	memcpy((void *)gd,(void *)good,FIELD_BYTES);
	
	for(i=1;i<=K_POWER-1;i++) {
		gf2x_mul(tmp,ai,gd,IRR_256);
		memcpy((void *)ai,(void *)tmp,FIELD_BYTES);
		memcpy((void *)xi,(void *)bad+i*FIELD_BYTES,FIELD_BYTES);
		gf2x_mul(tmp,xi,ai,IRR_256);
		gf2x_add(FIELD_LIMBS,acc,acc,tmp);
	}
	
	memcpy((void *)output,(void *)good,FIELD_BYTES);
	memcpy((void *)output+FIELD_BYTES,(void *)acc,FIELD_BYTES);
	
	MOD_DEC_USE_COUNT;
	
	return;
	
}


module_init(k1q_uhfe_init);
module_exit(k1q_uhfe_cleanup);
